/* ip_frag_sup.ex.c */
/* Fragmentation and Reassembly support procedures for the IP example model */

/****************************************/
/*	Copyright (c) 1987 - 1994			*/
/*			by MIL 3, Inc.				*/
/*		(A Delaware Corporation)		*/
/*	3400 International Drive,  N.W.		*/
/*		Washington, D.C., U.S.A.		*/
/*			All Rights Reserved.		*/
/****************************************/

#include <opnet.h>
#include "ip_frag_sup.h"



ip_dgram_list*
ip_frag_sup_setup ()
	{
	ip_dgram_list*		dl_ptr;

	/* Initialize a datagram list for an IP module. */
	FIN (ip_frag_sup_setup ())

	/* Create a new and empty list to keep track of datagrams. */
	dl_ptr = (ip_dgram_list*) op_prg_mem_alloc (sizeof (ip_dgram_list));
	dl_ptr->dgram_list_ptr = op_prg_list_create ();

	FRET (dl_ptr)
	}


Packet*
ip_frag_sup_insert (dl_ptr, pkptr)
	ip_dgram_list*		dl_ptr;
	Packet*				pkptr;
	{
	Packet*				ip_pkptr;
	int					len, total_len, ident, pk_id, trace_on = 0;
	int					dgram_list_size, src_net, src_node, i;
	ip_frag_file*		ifl_ptr;
	char				str0 [512], str1 [512];
	static	int			frag_file_defined = OPC_FALSE;
	static	Pmohandle	frag_file_pmh;


	/* Process a datagram fragment by inserting it into the fragment list. */
	/* If the fragment completes the arrival of a full ip datagram, return */
	/* the ip_datagram, otherwise return OPC_NIL. */
	FIN (ip_frag_sup_insert (dl_ptr, pkptr))

	if (op_prg_odb_ltrace_active ("ip_frag"))
		{
		/* Obtain the packet id; this wil be used */
		/* in subsequent debug statements as well. */
		pk_id = op_pk_id (pkptr);

		/* set a flag indicating that tracing is requested for */
		/* this reassembly operation. */
		trace_on = 1;	

		sprintf (str0, "insert pk (%d) into reassembly table", pk_id);
		op_prg_odb_print_major (str0, OPC_NIL);
		}

	/* Obtain the number of bytes contained in this datagram fragment. */
	op_pk_nfd_get (pkptr, "frag_len", &len);

	/* Obtain the total size of the encompassing datagram. */
	op_pk_nfd_get (pkptr, "orig_len", &total_len);

	/* First process the simplest case where the packet */
	/* is not fragmented to begin with. */
	if (len == total_len)
		{
		/* Simply return the datagram; no need to file it. */

		if (trace_on)
			{
			sprintf (str0, "returning unfragmented datagram: pk (%d)", pk_id);
			op_prg_odb_print_major (str0, OPC_NIL);
			}

		FRET (pkptr);
		}

	/* Obtain the identification of the encompasisng datagram. */
	op_pk_nfd_get (pkptr, "ident", &ident);

	/* Obtain the source address of the fragment. */
	op_pk_nfd_get (pkptr, "src_net", &src_net);
	op_pk_nfd_get (pkptr, "src_node", &src_node);

	/* Test to see if any other fragments of the same datagram have been received. */
	dgram_list_size = op_prg_list_size (dl_ptr->dgram_list_ptr);
	for (i = 0; i < dgram_list_size; i++)
		{
		/* Obtain the address of the i_th fragment list. */
		ifl_ptr = (ip_frag_file*) op_prg_list_access (dl_ptr->dgram_list_ptr, i);

		/* Test for an identification match */
		if (src_net == ifl_ptr->src_net &&
			src_node == ifl_ptr->src_node &&
			ident == ifl_ptr->dgram_id)
			{
	
			if (trace_on)
				{
				sprintf (str0, "fragment pk (%d) adds to previous frags", pk_id);
				op_prg_odb_print_major (str0, OPC_NIL);
				}

			/* Add in the length of the new fragment. */
			ifl_ptr->total_bytes_rcvd += len;

			/* If the new fragment is bearing the overall packet, */
			/* extract it and place it in the fragment list. */
			if (op_pk_nfd_is_set (pkptr, "ip_dgram"))
				{
				op_pk_nfd_get (pkptr, "ip_dgram", &ifl_ptr->ip_pkptr);
				}

			/* Discard the fragment (only its length is of interest) */
			op_pk_destroy (pkptr);

			/* Test if this completes the overall datagram. */
			if (ifl_ptr->total_bytes_rcvd == ifl_ptr->total_bytes_reqd)
				{
				if (trace_on)
					{
					sprintf (str0, "pk (%d) is completing fragment", pk_id);
					op_prg_odb_print_major (str0, OPC_NIL);
					}

				/* Extract the overall ip datagram so that it can be returned. */
				ip_pkptr = ifl_ptr->ip_pkptr;

				/* remove the ip fragment record from the datagram list. */
				op_prg_list_remove (dl_ptr->dgram_list_ptr, i);

				/* Deallocate its memory */
				op_prg_mem_free (ifl_ptr);
			
				/* Return the ip packet. */
				FRET (ip_pkptr)
				}
			else{
				FRET (OPC_NIL)
				}
			}
		}
		
	if (trace_on)
		{
		sprintf (str0, "fragment pk (%d) is first of several", pk_id);
		op_prg_odb_print_major (str0, OPC_NIL);
		}

	/* If no match was found, create a new fragment file to hold the fragment. 	*/
	/* Fragment files, which are used to reconstruct datagrams as 				*/
	/* component fragments arrive, are allocated as 'pooled' memory				*/
	/* objects. This is more efficient than ordinary dynamic memory				*/
	/* because these objects are frequently destroyed and reallocated			*/
	/* and can therefore be easily recycled. Defined the corresponding pooled	*/
	/* memory object type, if this has not already been done.					*/
	if (frag_file_defined == OPC_FALSE)
		{
		/* Define the object and request blocks of 32 objects to be allocated at a time. */
		frag_file_pmh = op_prg_pmo_define ("ip frag file", sizeof (ip_frag_file), 32);

		/* Prevent redefinition of the object. */
		frag_file_defined = OPC_TRUE;
		}

	/* Allocate the fragment.	*/
	ifl_ptr = (ip_frag_file*) op_prg_pmo_alloc (frag_file_pmh);

	/* Initialize fields of the fragment file. */
	ifl_ptr->dgram_id = ident;
	ifl_ptr->src_net = src_net;
	ifl_ptr->src_node = src_node;
	ifl_ptr->total_bytes_rcvd = len;
	ifl_ptr->total_bytes_reqd = total_len;
	ifl_ptr->ip_pkptr = OPC_NIL;

	/* If this first fragment contains the overall ip packet, extract it */
	if (op_pk_nfd_is_set (pkptr, "ip_dgram"))
		{
		op_pk_nfd_get (pkptr, "ip_dgram", &ifl_ptr->ip_pkptr);
		}

	/* Append the fragment file to the list */
	op_prg_list_insert (dl_ptr->dgram_list_ptr, ifl_ptr, OPC_LISTPOS_TAIL);

	/* Discard the ip fragment. */
	op_pk_destroy (pkptr);

	/* Return NIL to indicate that the datagram is not complete yet. */
	FRET (OPC_NIL)
	}
	


ip_frag_sup_print (dl_ptr)
	ip_dgram_list*		dl_ptr;
	{
	ip_frag_file*		f_ptr;
	int					i;
	char				str0 [128], str1 [128], str2 [128];


	FIN (ip_frag_sup_print (dl_ptr))

	/* If this procedure is called and the datagram list has */
	/* not yet been setup, simply return. */
	if (dl_ptr == VOS_NIL)
		FOUT

	for (i = 0; i < op_prg_list_size (dl_ptr->dgram_list_ptr); i++)
		{
		f_ptr = (ip_frag_file*) op_prg_list_access (dl_ptr->dgram_list_ptr, i);

		sprintf (str0, "entry %d: src (%d, %d) id (%d)", 
					i, f_ptr->src_net, f_ptr->src_node, f_ptr->dgram_id);
		sprintf (str1, "bytes_rcvd (%d), bytes_expected (%d)", 
					f_ptr->total_bytes_rcvd, f_ptr->total_bytes_reqd);
		sprintf (str2, "encap packet rcvd: %s", 
				(f_ptr->ip_pkptr == OPC_NIL) ? "NO" : "YES");
		op_prg_odb_print_major (str0, str1, str2, OPC_NIL);
		}


	FOUT
	}









